// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) ZenJu (zhnmju123 AT gmx DOT de) - All Rights Reserved    *
// **************************************************************************

#include "unittest.h"
#include "../resources.h"
#include <zen/file_handling.h>
#include "../resolve_path.h"
#include "../hard_filter.h"
#include "../../algorithm.h"
#include <wx+/format_unit.h>
#include <wx+/string_conv.h>

using namespace zen;


/*
TEST(Miscellaneous, RemoveRowsFromVector)
{
    std::vector<int> test; //0 1 2 3 4 5 6
    for (int i = 0; i < 7; ++i)
        test.push_back(i);

    std::set<unsigned int> rowsToRemove;
    rowsToRemove.insert(-41); //invalid
    rowsToRemove.insert(1);
    rowsToRemove.insert(2);
    rowsToRemove.insert(5);
    rowsToRemove.insert(15);  //invalid
    rowsToRemove.insert(100); //invalid

    common::removeRowsFromVector(rowsToRemove, test);
    //0 3 4 6

    ASSERT_EQ(unsigned(4), test.size());
    ASSERT_EQ(0, test[0]);
    ASSERT_EQ(3, test[1]);
    ASSERT_EQ(4, test[2]);
    ASSERT_EQ(6, test[3]);
}
*/

TEST(Formatting, DirectoryNames)
{
#ifdef FFS_WIN
    ASSERT_PRED_FORMAT2(assertEqual, toWx(zen::getFormattedDirectoryName(Zstr("  c:\\test "))), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, toWx(zen::getFormattedDirectoryName(Zstr("  c:\\test\\ "))), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, toWx(zen::getFormattedDirectoryName(Zstr("  \\abc\\ "))), wxT("C:\\abc\\"));
    ASSERT_PRED_FORMAT2(assertEqual, toWx(zen::getFormattedDirectoryName(Zstr("  \\a\\b\\c "))), wxT("C:\\a\\b\\c\\"));
    ASSERT_PRED_FORMAT2(assertEqual, toWx(zen::getFormattedDirectoryName(Zstr("  \\a\\b\\c d  "))), wxT("C:\\a\\b\\c d\\"));
    ASSERT_PRED_FORMAT2(assertEqual, toWx(zen::getFormattedDirectoryName(Zstr("  \\\\a b/b\\c d  "))), wxT("\\\\a b\\b\\c d\\"));
#endif
}


//predicate-formatter for asserting that two vector<Zstring> are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const std::vector<Zstring>& a, const std::vector<Zstring>& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = a.begin(); i != a.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " and " << b_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = b.begin(); i != b.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " not equal!";

    return testing::AssertionFailure(msg);
}

std::vector<Zstring> splitByDelimiter(const Zstring& filterString);

TEST(Formatting, FilterString)
{
    std::vector<Zstring> filterActual = splitByDelimiter(Zstr("   A\n   b;; c;\n\n;\n d    ;E    f  \n;& ;g\\"));
    std::vector<Zstring> filterExpected;
    filterExpected.push_back(Zstr("A"));
    filterExpected.push_back(Zstr("b"));
    filterExpected.push_back(Zstr("c"));
    filterExpected.push_back(Zstr("d"));
    filterExpected.push_back(Zstr("E    f"));
    filterExpected.push_back(Zstr("&"));
    filterExpected.push_back(Zstr("g\\"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = splitByDelimiter(Zstr("   ; ;\n a"));
    filterExpected.clear();
    filterExpected.push_back(Zstr("a"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = splitByDelimiter(Zstr("   ; ;\n       ;\n\n"));
    filterExpected.clear();
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);
}


TEST(Formatting, FileSize)
{
    wxString sizeActual = zen::filesizeToShortString(807);
    wxString sizeExpected = _P("1 Byte", "%x Bytes", 807);

    sizeExpected.Replace(wxT("%x"), wxT("807"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = zen::filesizeToShortString(315621376);
    sizeExpected = _("%x MB");
    sizeExpected.Replace(wxT("%x"), wxT("301"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = zen::filesizeToShortString(48954);
    sizeExpected = _("%x KB");
    sizeExpected.Replace(wxT("%x"), wxT("47.8"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = zen::filesizeToShortString(35032904);
    sizeExpected = _("%x MB");
    sizeExpected.Replace(wxT("%x"), wxT("33.4"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = zen::filesizeToShortString(Int64(3084582912U));
    sizeExpected = _("%x GB");
    sizeExpected.Replace(wxT("%x"), wxT("2.87"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = zen::filesizeToShortString(1001);
    sizeExpected = _("%x KB");
    sizeExpected.Replace(wxT("%x"), wxT("0.98"));
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);
}


TEST(Filtering, Standard)
{
#ifdef FFS_WIN
    zen::NameFilter filterIncl(Zstr("*abc?5*2  ;    abc?3;   name\\; test\\subdir "), Zstr(""));
#elif defined FFS_LINUX
    zen::NameFilter filterIncl(Zstr("*abc?5*2  ;    abc?3;   name/; test/subdir "), Zstr(""));
#endif

    ASSERT_TRUE(filterIncl.passFileFilter( Zstr("zzzzabcz5zzzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(Zstr("zzzzabczz5zzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(Zstr("zzzzabcz5zzzzz")));

    ASSERT_TRUE(filterIncl.passFileFilter(Zstr("abcz3")));
    ASSERT_FALSE(filterIncl.passFileFilter(Zstr("abcz")));

    bool continueTrav = true;
    ASSERT_TRUE(filterIncl.passDirFilter(Zstr("name"), &continueTrav));

    ASSERT_FALSE(filterIncl.passDirFilter(Zstr("test"), &continueTrav));
    ASSERT_TRUE(continueTrav); //but continue traversing, because subdir might match filter
    //------------------------------------------------------------------------------------------

#ifdef FFS_WIN
    zen::NameFilter filterExcl(Zstr("*"), Zstr("asdf?5 ;;; name\\*"));
#elif defined FFS_LINUX
    zen::NameFilter filterExcl(Zstr("*"), Zstr("asdf?5 ;;; name/*"));
#endif

    ASSERT_FALSE(filterExcl.passFileFilter(Zstr("asdfz5")));
    ASSERT_TRUE(filterExcl.passFileFilter(Zstr("asdfz")));

    ASSERT_FALSE(filterExcl.passDirFilter(Zstr("name"), &continueTrav));
    ASSERT_FALSE(continueTrav);
    //------------------------------------------------------------------------------------------

    {
        zen::NameFilter filter(Zstr("abc*?def"), Zstr("")); //*? -pattern: 1..n chars

        ASSERT_TRUE (filter.passFileFilter(Zstr("abcXdef")));
        ASSERT_FALSE(filter.passFileFilter(Zstr("abcdef")));
        ASSERT_TRUE (filter.passFileFilter(Zstr("abcXXdef")));
    }
}


TEST(Zstring, AfterBefore)
{
    ASSERT_EQ(afterLast(Zstring(Zstr("a/dsfsadf/abc")), Zchar('/')), Zstring(Zstr("abc")));
    ASSERT_EQ(beforeLast(Zstring(Zstr("a/dsfsadf/abc")), Zchar('/')), Zstring(Zstr("a/dsfsadf")));
    ASSERT_EQ(afterFirst(Zstring(Zstr("a/dsfsadf/abc")), Zchar('/')), Zstring(Zstr("dsfsadf/abc")));
    ASSERT_EQ(beforeFirst(Zstring(Zstr("a/dsfsadf/abc")), Zchar('/')), Zstring(Zstr("a")));
}


TEST(Zstring, ReverseFind)
{
    Zstring phrase(Zstr("----abc-"));
    const size_t targetPos = 4;

    ASSERT_EQ(phrase.rfind(Zstr("abc")),    targetPos);
    ASSERT_EQ(phrase.rfind(Zstr("abc"), 7), targetPos);

    ASSERT_EQ(phrase.rfind(Zstr("abc"), 3), size_t(Zstring::npos));
    ASSERT_EQ(phrase.rfind(Zstr("abc"), 4), targetPos);
    ASSERT_EQ(phrase.rfind(Zstr("abc"), 5), targetPos);
}


TEST(Zstring, Resize)
{
    Zstring testStr(Zstr("abcdefgh"));
    testStr.resize(3);
    ASSERT_EQ(testStr, Zstring(Zstr("abc")));

    Zstring testStr2(Zstr("abcdefgh"));
    testStr2.resize(10); //add 0-characters!
    ASSERT_EQ(testStr2, Zstring(Zstr("abcdefgh\0\0"), 10));

    Zstring testStr3(Zstr("abcdefgh"));
    testStr3.resize(3, Zchar('N'));
    ASSERT_EQ(testStr3, Zstring(Zstr("abc")));

    Zstring testStr4(Zstr("abcdefgh"));
    testStr4.resize(10, Zchar('N')); //add 'N' characters!
    ASSERT_EQ(testStr4, Zstring(Zstr("abcdefghNN")));
}


TEST(String_Tools, Split)
{
    std::vector<std::string> result;
    std::vector<std::string> expected;

    std::string testStr = ";a; b;c ;d ;";
    result = split(testStr, ";");

    expected.push_back("");
    expected.push_back("a");
    expected.push_back(" b");
    expected.push_back("c ");
    expected.push_back("d ");
    expected.push_back("");

    ASSERT_EQ(result, expected);

    //------------------------------------------------
    testStr = "--a-- b---c --d-- ";
    result  = split(testStr, "--");

    expected.clear();
    expected.push_back("");
    expected.push_back("a");
    expected.push_back(" b");
    expected.push_back("-c ");
    expected.push_back("d");
    expected.push_back(" ");

    ASSERT_EQ(result, expected);

    //------------------------------------------------
    testStr = "xxx";
    result  = split(testStr, "xxx");

    expected.clear();
    expected.push_back("");
    expected.push_back("");

    ASSERT_EQ(result, expected);

    //------------------------------------------------
    testStr = "xx";
    result  = split(testStr, "xxx");

    expected.clear();
    expected.push_back("xx");

    ASSERT_EQ(result, expected);
}


TEST(String_Tools, ReverseSearch)
{
    //                    012345678
    std::string str    = "ababababa";

    std::string needle = "a";
    ASSERT_EQ(search_last(str.begin(), str.end(),
                          needle.begin(), needle.end()) - str.begin(), 8);

    needle = "c";
    ASSERT_EQ(search_last(str.begin(), str.end(),
                          needle.begin(), needle.end()) - str.begin(), 9);

    needle = "ab";
    ASSERT_EQ(search_last(str.begin(), str.end(),
                          needle.begin(), needle.end()) - str.begin(), 6);

    needle = "";
    ASSERT_EQ(search_last(str.begin(), str.end(),
                          needle.begin(), needle.end()) - str.begin(), 9);

    str    = "";
    needle = "jhjh";
    ASSERT_EQ(search_last(str.begin(), str.end(),
                          needle.begin(), needle.end()) - str.begin(), 0);

    str    = "";
    needle = "";
    ASSERT_EQ(search_last(str.begin(), str.end(),
                          needle.begin(), needle.end()) - str.begin(), 0);
}


TEST(STL_Test, nthElement)
{
    const int input[] = { 5, 4, 3, 2, 1 };
    std::vector<int> v(std::begin(input), std::end(input));

    std::nth_element(v.begin(), v.begin(), v.end());
    std::nth_element(v.begin(), v.end(), v.end()); //STL documentation is unclear on whether this is supported
}

