#ifndef LAUNCH_H_INCLUDED
#define LAUNCH_H_INCLUDED

#include <algorithm>
#include <string>
#include <vector>
#include <zen/dll.h>
#include <zen/build_info.h>
#include <zen/assert_static.h>
#include <zen/scope_guard.h>
#include <zen/string_tools.h>

#define WIN32_LEAN_AND_MEAN
#include <zen/win.h> //includes windows.h


bool runningWOW64() //test if process is running under WOW64 (reference http://msdn.microsoft.com/en-us/library/ms684139(VS.85).aspx)
{
    //dynamically load windows API function
    typedef BOOL (WINAPI* IsWow64ProcessFun)(HANDLE hProcess, PBOOL Wow64Process);
    const zen::SysDllFun<IsWow64ProcessFun> isWow64Process(L"kernel32.dll", "IsWow64Process");
    if (isWow64Process)
    {
        BOOL isWow64 = FALSE;
        if (isWow64Process(::GetCurrentProcess(), &isWow64))
            return isWow64 == TRUE;
    }
    return false;
}



int launch(const std::wstring& exe32bit, const std::wstring& exe64bit) //returns called process' return code
{
    //get unicode commandline
    std::wstring newCmdLine;
    {
        std::vector<std::wstring> argv;
        int argc = 0;
        if (LPWSTR* tmp = ::CommandLineToArgvW(::GetCommandLine(), &argc))
        {
            std::copy(tmp, tmp + argc, std::back_inserter(argv));
            ::LocalFree(tmp);
        }

        if (!argv.empty())
        {
            //replace executable name
            size_t pos = argv[0].rfind(L'\\');
            if (pos != std::wstring::npos)
                newCmdLine = argv[0].substr(0, pos + 1);

            assert_static(zen::is32BitBuild);
            newCmdLine += runningWOW64() ? exe64bit : exe32bit;
            newCmdLine = L'\"' + newCmdLine + L'\"';

            std::for_each(argv.begin() + 1, argv.end(),
                          [&](const std::wstring& arg)
            {
                if (std::any_of(arg.begin(), arg.end(), [](wchar_t c) { return zen::isWhiteSpace(c); }))
                {
                    if (zen::endsWith(arg, L'\\'))
                        newCmdLine += L" \"" + arg + L" \"";
                    //attention: \" is parsed differently: E.g. "C:\" "d:\"  -> C:\" d:\"
                    //therefore place a space before last quotation mark! However this may *corrupt* the command line so do this only when absolutely required!
                    //http://msdn.microsoft.com/en-us/library/windows/desktop/17w5ykft(v=vs.85).aspx
                    else
                        newCmdLine += L" \"" + arg + L"\"";
                }
                else
                    newCmdLine += L" " + arg;
            });
        }
    }

    //::MessageBox(0, redCmdLine.c_str(), L"Info", 0);

    STARTUPINFO startupInfo = {};
    startupInfo.cb = sizeof(startupInfo);
    //startupInfo.dwFlags = STARTF_FORCEOFFFEEDBACK; //avoid showing busy cursor when starting process

    PROCESS_INFORMATION procInfo = {};
    if (!::CreateProcess(nullptr,        //__in_opt     LPCTSTR lpApplicationName,
                         &newCmdLine[0], //__inout_opt  LPTSTR lpCommandLine,
                         nullptr,        //__in_opt     LPSECURITY_ATTRIBUTES lpProcessAttributes,
                         nullptr,        //__in_opt     LPSECURITY_ATTRIBUTES lpThreadAttributes,
                         TRUE,           //__in         BOOL bInheritHandles,
                         0,              //__in         DWORD dwCreationFlags,
                         nullptr,        //__in_opt     LPVOID lpEnvironment,
                         nullptr,        //__in_opt     LPCTSTR lpCurrentDirectory,
                         &startupInfo,   //__in         LPSTARTUPINFO lpStartupInfo,
                         &procInfo))     //__out        LPPROCESS_INFORMATION lpProcessInformation
    {
        ::MessageBox(0, (L"Error launching: " + newCmdLine).c_str(), L"Error", 0);
        return -1;
    }

    ZEN_ON_SCOPE_EXIT(::CloseHandle(procInfo.hProcess));
    ZEN_ON_SCOPE_EXIT(::CloseHandle(procInfo.hThread));

    //synchronous call
    //::WaitForSingleObject(procInfo.hProcess, INFINITE); -> if we don't process the message loop, Windows will show the nasty busy cursor!
    [&]
    {
        for (;;)
        {
            MSG msg = {};
            const DWORD rv = ::MsgWaitForMultipleObjects(1, &procInfo.hProcess, FALSE, INFINITE, QS_ALLINPUT); //wait for child process or *new* messages
            switch (rv)
            {
                case WAIT_OBJECT_0: //child process has finished
                    return;
                case WAIT_OBJECT_0 + 1: //new message available
                    while (::PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
                        //attention: a single call to ::PeekMessage() marks *all* waiting messages as old
                        //with regards to ::MsgWaitForMultipleObjects! => always clear complete message queue!
                    {
                        if (msg.message == WM_QUIT)
                            return;
                        ::TranslateMessage(&msg);
                        ::DispatchMessage(&msg);
                    }
                    break;
                case WAIT_ABANDONED_0: //
                case WAIT_TIMEOUT:     // not expected
                case WAIT_FAILED:      //
                default:               //
                    return;
            }
        }
    }();

    //redirect return code
    DWORD exitCode = 0;
    if (!::GetExitCodeProcess(procInfo.hProcess, // __in   HANDLE hProcess,
                              &exitCode) ||      //__out  LPDWORD lpExitCode
        exitCode == STILL_ACTIVE)
        exitCode = -100;

    return exitCode;
}

#endif // LAUNCH_H_INCLUDED
